from .base import UdgerBase

from .queries import Queries


class UaRequest(object):
    def __init__(self, ua_string=None, sec_ch_ua=None, sec_ch_ua_full_version_list=None, sec_ch_ua_mobile=None,
                 sec_ch_ua_full_version=None, sec_ch_ua_platform=None, sec_ch_ua_platform_version=None, sec_ch_ua_model=None):
        self.ua_string = ua_string
        self.sec_ch_ua = sec_ch_ua
        self.sec_ch_ua_full_version_list = sec_ch_ua_full_version_list.strip('"') if sec_ch_ua_full_version_list else None
        self.sec_ch_ua_mobile = sec_ch_ua_mobile
        self.sec_ch_ua_full_version = sec_ch_ua_full_version.strip('"') if sec_ch_ua_full_version else None
        self.sec_ch_ua_platform = sec_ch_ua_platform.strip('"') if sec_ch_ua_platform else None
        self.sec_ch_ua_platform_version = sec_ch_ua_platform_version.strip('"') if sec_ch_ua_platform_version else None
        self.sec_ch_ua_model = sec_ch_ua_model.strip('"') if sec_ch_ua_model else None

    def cache_key(self):
        return '[uaString=' + (self.ua_string if self.ua_string is not None else '') + \
                ', secChUa=' + (self.sec_ch_ua if self.sec_ch_ua is not None else '') + \
                ', secChUaFullVersionList=' + (self.sec_ch_ua_full_version_list if self.sec_ch_ua_full_version_list is not None else '') + \
                ', secChUaMobile=' + (self.sec_ch_ua_mobile if self.sec_ch_ua_mobile is not None else '') + \
                ', secChUaFullVersion=' + (self.sec_ch_ua_full_version if self.sec_ch_ua_full_version is not None else '') + \
                ', secChUaPlatform=' + (self.sec_ch_ua_platform if self.sec_ch_ua_platform is not None else '') + \
                ', secChUaPlatformVersion=' + (self.sec_ch_ua_platform_version if self.sec_ch_ua_platform_version is not None else '') + \
                ', secChUaModel=' + (self.sec_ch_ua_model if self.sec_ch_ua_model is not None else '') + ']'


class Udger(UdgerBase):

    def parse_ua(self, ua_string):
        ua_request = UaRequest(ua_string=ua_string)
        return self.parse_ua_request(ua_request)

    def parse_ua_request(self, ua_request):

        cache_key = None
        if self.lru_cache is not None and ua_request is not None:
            cache_key = ua_request.cache_key()
            cached = self.lru_cache.get(cache_key, None)
            if cached is not None:
                return cached

        ua, class_id, client_id = self._process_client(ua_request.ua_string)
        is_crawler = (ua['ua_class'] == 'Crawler')

        opsys = self._process_os(ua_request.ua_string, client_id) if not is_crawler else None
        ua.update(opsys or self.os_emptyrow)

        dev = self._process_device(ua_request.ua_string, class_id)
        ua.update(dev or self.device_emptyrow)

        marketname = self._process_marketname(ua, ua_request.ua_string, class_id) if not is_crawler and ua['os_family_code'] else None
        ua.update(marketname or self.marketname_emptyrow)

        if ua.get('ua_class_code') != 'Crawler' and self.client_hints_parser_enabled:
            ua.update(vars(ua_request))
            self._process_client_hints(ua, ua_request)
        else:
            ua['ua_string'] = ua_request.ua_string

        if cache_key is not None:
            self.lru_cache[cache_key] = ua

        return ua

    def parse_ip(self, ip_string):
        ip = self.ip_datacenter_emptyrow.copy()
        ip['ip'] = ip_string

        try:
            ip_string, ipv4_int, ipv6_words = self.normalize_ipaddress(ip_string)
        except:
            pass
        else:
            ip.update(
                ip_classification='Unrecognized',
                ip_classification_code='unrecognized',
            )

            ip_row = self.db_get_first_row(Queries.ip_sql, ip_string)
            if ip_row:
                if ip_row['ip_classification_code'] != 'crawler':
                    ip_row.pop('crawler_family_info_url')

                ip.update(ip_row)

            if ipv4_int is not None:
                ip['ip_ver'] = 4
                dc = self.db_get_first_row(Queries.datacenter_sql, ipv4_int, ipv4_int)

            else:
                ip['ip_ver'] = 6
                ipv6_words *= 2
                dc = self.db_get_first_row(Queries.datacenter6_sql, *ipv6_words)

            if dc:
                ip.update(dc)

        return ip

    def _process_client(self, ua_string):
        if not ua_string:
            ua = self.client_empty_row.copy()
            class_id = ua.pop('class_id', -1)
            client_id = ua.pop('client_id', 0)
        else:
            ua = self.db_get_first_row(Queries.crawler_sql, ua_string)
            if ua:
                del ua['class_id']
                del ua['client_id']
                class_id = 99
                client_id = -1
            else:
                rowid = self._find_id_from_list(ua_string, self.client_word_detector.find_words(ua_string), self.client_regstring_list)
                if rowid != -1:
                    ua = self.db_get_first_row(Queries.client_sql, rowid)
                    self._patch_versions(ua)
                else:
                    ua = self.client_unrecognized_row.copy()
                class_id = ua.pop('class_id', -1)
                client_id = ua.pop('client_id', 0)
        return ua, class_id, client_id

    def _process_os(self, ua_string, client_id):
        rowid = self._find_id_from_list(ua_string, self.os_word_detector.find_words(ua_string), self.os_regstring_list)
        if rowid != -1:
            return self.db_get_first_row(Queries.os_sql, rowid)
        return client_id != 0 and self.db_get_first_row(Queries.client_os_sql, client_id)

    def _process_device(self, ua_string, class_id):
        rowid = self._find_id_from_list(ua_string, self.device_word_detector.find_words(ua_string), self.device_regstring_list)
        if rowid != -1:
            return self.db_get_first_row(Queries.device_sql, rowid)
        return class_id != -1 and self.db_get_first_row(Queries.client_class_sql, class_id)

    def _process_marketname(self, ua, ua_string, class_id):
        marketname = None
        # must complete first so cursors don't collide
        rows = tuple(self.db_iter_rows(
            Queries.devicename_sql,
            ua['os_family_code'],
            ua['os_code'],
        ))

        for dn_row in rows:
            if self.regexp_func(dn_row['regstring'], ua_string):
                match = self.last_regexp_match.group(1)

                marketname = self.db_get_first_row(
                    Queries.marketname_sql,
                    dn_row['regex_id'],
                    match.strip(),
                )
                if marketname:
                    break
        return marketname

    def _process_client_hints(self, ua, ua_request):

        if not ua_request.sec_ch_ua_mobile:
            sec_ch_ua_mobile = 0
            ua['sec_ch_ua_mobile'] = ''
        else:
            if ua_request.sec_ch_ua_mobile == '?0':
                sec_ch_ua_mobile = 0
            else:
                sec_ch_ua_mobile = 1
            ua['sec_ch_ua_mobile'] = sec_ch_ua_mobile

        regstring_search1 = ua_request.sec_ch_ua_full_version_list

        if not regstring_search1:
            regstring_search1 = ua_request.sec_ch_ua

        if regstring_search1:
            rows1 = tuple(self.db_iter_rows(Queries.client_ch_regex_sql, sec_ch_ua_mobile))

            for dn_row in rows1:
                regex = dn_row['regstring']
                if regex and self.regexp_func(regex, regstring_search1):
                    ver = self.last_regexp_match.group(1)

                    if ua_request.sec_ch_ua_full_version_list:
                        dot_index = ver.find('.')
                        ver_major = ver[0:dot_index] if dot_index >= 0 else ver
                    else:
                        ver_major = ver
                        if ua_request.sec_ch_ua_full_version:
                            ver = ua_request.sec_ch_ua_full_version

                    ua['ua_class'] = dn_row['client_classification']
                    ua['ua_class_code'] = dn_row['client_classification_code']
                    ua['ua'] = dn_row['name'] + ' ' + ver
                    ua['ua_version'] = ver
                    ua['ua_version_major'] = ver_major
                    ua['ua_uptodate_current_version'] = dn_row['uptodate_current_version']
                    ua['ua_family'] = dn_row['name']
                    ua['ua_family_code'] = dn_row['name_code']
                    ua['ua_family_homepage'] = dn_row['homepage']
                    ua['ua_family_vendor'] = dn_row['vendor']
                    ua['ua_family_vendor_code'] = dn_row['vendor_code']
                    ua['ua_family_vendor_homepage'] = dn_row['vendor_homepage']
                    ua['ua_family_icon'] = dn_row['icon']
                    ua['ua_family_icon_big'] = dn_row['icon_big']
                    ua['ua_family_info_url'] = dn_row['ua_family_info_url']
                    ua['ua_engine'] = dn_row['engine']
                    break

        regstring_search2 = ua_request.sec_ch_ua_platform
        if regstring_search2:
            rows2 = tuple(self.db_iter_rows(Queries.os_ch_regex_sql,
                                            ua_request.sec_ch_ua_platform_version if ua_request.sec_ch_ua_platform_version else ''))

            for dn_row in rows2:
                    regex = dn_row['regstring']
                    if regex and self.regexp_func(regex, regstring_search2):
                        ua['os'] = dn_row['name']
                        ua['os_code'] = dn_row['name_code']
                        ua['os_homepage'] = dn_row['homepage']
                        ua['os_icon'] = dn_row['icon']
                        ua['os_icon_big'] = dn_row['icon_big']
                        ua['os_info_url'] = dn_row['os_info_url']
                        ua['os_family'] = dn_row['family']
                        ua['os_family_code'] = dn_row['family_code']
                        ua['os_family_vendor'] = dn_row['vendor']
                        ua['os_family_vendor_code'] = dn_row['vendor_code']
                        ua['os_family_vendor_homepage'] = dn_row['vendor_homepage']
                        break

        if ua_request.sec_ch_ua_model and ua['os_family_code']:
            row3 = self.db_get_first_row(Queries.device_name_ch_regex_sql, ua['os_family_code'], ua['os_family_code'], ua['os_code'])
            if row3:
                row4 = self.db_get_first_row(Queries.device_name_list_ch_sql, row3['id'], ua_request.sec_ch_ua_model)
                if row4:
                    ua['device_marketname'] = row4['marketname']
                    ua['device_brand'] = row4['brand']
                    ua['device_brand_code'] = row4['brand_code']
                    ua['device_brand_homepage'] = row4['brand_url']
                    ua['device_brand_icon'] = row4['icon']
                    ua['device_brand_icon_big'] = row4['icon_big']
                    ua['device_brand_info_url'] = row4['brand_info_url']

                    row5 = self.db_get_first_row(Queries.device_class_ch_sql, row4['deviceclass_id'])
                    if row5:
                        ua['device_class'] = row5['device_class']
                        ua['device_class_code'] = row5['device_class_code']
                        ua['device_class_icon'] = row5['device_class_icon']
                        ua['device_class_icon_big'] = row5['device_class_icon_big']
                        ua['device_class_info_url'] = row5['device_class_info_url']


        if not ua['device_class'] and ua['ua_class_code']:
            row6 = self.db_get_first_row(Queries.device_class_by_mobile_ch_sql, sec_ch_ua_mobile)
            if row6:
                    ua['device_class'] = row6['device_class']
                    ua['device_class_code'] = row6['device_class_code']
                    ua['device_class_icon'] = row6['device_class_icon']
                    ua['device_class_icon_big'] = row6['device_class_icon_big']
                    ua['device_class_info_url'] = row6['device_class_info_url']

    def _find_id_from_list(self, ua_string, found_client_words, reg_string_list):
        self.last_regexp_match = None
        for irs in reg_string_list:
            if (irs.word_id in found_client_words) and (irs.word2_id in found_client_words):
                m = irs.pattern.search(ua_string)
                if m is not None:
                    self.last_regexp_match = m
                    return irs.rowid
        return -1

    def _patch_versions(self, ua):
        if self.last_regexp_match:
            try:
                ver = self.last_regexp_match.group(1)
            except IndexError:
                ver = ''

            ua['ua_version'] = ver
            ua['ua'] += ' ' + ver
            ua['ua_version_major'] = ver.split('.')[0]
        else:
            ua['ua_version'] = ua['ua_version_major'] = None




import os
import re
import sys
import socket
import sqlite3
import struct
import tempfile

from .queries import Queries
from .wdetector import *


if sys.version_info.major == 3 and sys.version_info.minor >= 10:
    from collections.abc import MutableMapping
else:
    from collections import MutableMapping

unperlize_re = re.compile('^/?(.*)/([si]*)$')


def make_empty_row(columns_dict):
    return dict((col, None) for col in columns_dict)


class cached_property(object):
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.
    """
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, unused_type=None):
        if instance is None:
            return self
        res = instance.__dict__[self.func.__name__] = self.func(instance)
        return res


class LRUDict(MutableMapping):
    def __init__(self, maxlen, *a, **k):
        self.maxlen = maxlen
        self.d = dict(*a, **k)
        while len(self) > maxlen:
            self.popitem()

    def __iter__(self):
        return iter(self.d)

    def __len__(self):
        return len(self.d)

    def __getitem__(self, k):
        return self.d[k]

    def __delitem__(self, k):
        del self.d[k]

    def __setitem__(self, k, v):
        if k not in self and len(self) == self.maxlen:
            self.popitem()
        self.d[k] = v


class IdRegString(object):
    def __init__(self, rowid, word_id, word2_id, pattern):
        self.rowid = rowid
        self.word_id = word_id
        self.word2_id = word2_id
        self.pattern = pattern


class UdgerBase(object):
    db_filename = 'udgerdb_v4.dat'

    _client_word_detector = None
    _os_word_detector = None
    _device_word_detector = None

    _client_regstring_list = None
    _os_regstring_list = None
    _device_regstring_list = None

    client_unrecognized_row = make_empty_row(Queries.client_columns)
    client_unrecognized_row.update(
        ua_class="Unrecognized",
        ua_class_code="unrecognized",
    )

    client_empty_row = make_empty_row(Queries.client_columns)
    client_empty_row.update(
        ua_class="",
        ua_class_code="",
    )

    os_emptyrow = make_empty_row(Queries.os_columns)
    device_emptyrow = make_empty_row(Queries.device_columns)
    marketname_emptyrow = make_empty_row(Queries.marketname_columns)

    ip_datacenter_emptyrow = make_empty_row(Queries.ip_columns)
    ip_datacenter_emptyrow.update(
        make_empty_row(Queries.datacenter_columns),
        ip_ver=None,
    )

    def __init__(self, client_hints_parser_enabled=True, data_dir=None, lru_cache_size=10000):
        self.client_hints_parser_enabled = client_hints_parser_enabled
        self.data_dir = data_dir or tempfile.gettempdir()
        self.regexp_cache = {}
        if lru_cache_size > 0:
            self.lru_cache = LRUDict(lru_cache_size)

    @staticmethod
    def dict_factory(cursor, row):
        return dict(
            (col[0], row[idx])
            for idx, col in enumerate(cursor.description)
        )

    perl_flags = {
        's': re.DOTALL,
        'i': re.IGNORECASE,
        'm': re.MULTILINE,
        'x': re.VERBOSE,
    }

    def regexp_func(self, expr, item):
        global unperlize_re

        expr_re = self.regexp_cache.get(expr)
        if expr_re is None:
            m = unperlize_re.match(expr)
            old_expr = expr
            flags = 0
            if m:
                # strip / from the beginning and /(si...) from the end
                expr, opts = m.groups()
                # this fails for unsupported Perl flag
                flags = sum(map(self.perl_flags.get, opts))

            expr_re = re.compile(expr, flags)
            self.regexp_cache[old_expr] = expr_re

        self.last_regexp_match = expr_re.search(item)  # this does not take flags!

        return bool(self.last_regexp_match)

    @cached_property
    def db_cursor(self):
        db_filepath = os.path.join(self.data_dir, self.db_filename)
        db = sqlite3.connect(db_filepath)
#        db.create_function("REGEXP", 2, self.regexp_func)

        cursor = db.cursor()
        cursor.row_factory = self.dict_factory

        return cursor

    @cached_property
    def client_regstring_list(self):
        if UdgerBase._client_regstring_list is None:
            UdgerBase._client_regstring_list = UdgerBase.prepare_regexp_struct(self.db_cursor, "udger_client_regex")
        return UdgerBase._client_regstring_list

    @cached_property
    def os_regstring_list(self):
        if UdgerBase._os_regstring_list is None:
            UdgerBase._os_regstring_list = UdgerBase.prepare_regexp_struct(self.db_cursor, "udger_os_regex")
        return UdgerBase._os_regstring_list

    @cached_property
    def device_regstring_list(self):
        if UdgerBase._device_regstring_list is None:
            UdgerBase._device_regstring_list = UdgerBase.prepare_regexp_struct(self.db_cursor, "udger_deviceclass_regex")
        return UdgerBase._device_regstring_list

    @staticmethod
    def prepare_regexp_struct(db_cursor, regexp_table_name):
        result = []
        global unperlize_re
        for row in db_cursor.execute('SELECT rowid, regstring, word_id, word2_id FROM %s ORDER BY sequence' % regexp_table_name):
            regstring = row['regstring']
            m = unperlize_re.match(regstring)
            flags = 0
            if m:
                # strip / from the beginning and /(si...) from the end
                regstring, opts = m.groups()
                # this fails for unsupported Perl flag
                flags = sum(map(UdgerBase.perl_flags.get, opts))
            expr_re = re.compile(regstring, flags)
            rs = IdRegString(row['rowid'], row['word_id'], row['word2_id'], expr_re)
            result.append(rs)

        return result;

    @cached_property
    def client_word_detector(self):
        if UdgerBase._client_word_detector is None:
            UdgerBase._client_word_detector = UdgerBase.create_word_detector(self.db_cursor, 'udger_client_regex', 'udger_client_regex_words')
        return UdgerBase._client_word_detector

    @cached_property
    def device_word_detector(self):
        if UdgerBase._device_word_detector is None:
            UdgerBase._device_word_detector = UdgerBase.create_word_detector(self.db_cursor, 'udger_deviceclass_regex', 'udger_deviceclass_regex_words')
        return UdgerBase._device_word_detector

    @cached_property
    def os_word_detector(self):
        if UdgerBase._os_word_detector is None:
            UdgerBase._os_word_detector = UdgerBase.create_word_detector(self.db_cursor, 'udger_os_regex', 'udger_os_regex_words')
        return UdgerBase._os_word_detector

    @staticmethod
    def create_word_detector(db_cursor, regex_table, word_table_name):
        wdetector = WordDetector()
        sql = "SELECT %s FROM " + regex_table
        used_words = set(row['word_id'] for row in db_cursor.execute(sql % 'word_id'))
        used_words |= set(row['word2_id'] for row in db_cursor.execute(sql % 'word2_id'))

        for row in db_cursor.execute('SELECT * FROM %s' % word_table_name):
            if row['id'] in used_words:
                wdetector.add_word(row['id'], row['word'])

        return wdetector

    def db_get_first_row(self, sql, *params):
#        self.last_regexp_match = None

        self.db_cursor.execute(sql, params)

        for row in self.db_cursor:
            return row

    def db_iter_rows(self, sql, *params):
#        self.last_regexp_match = None

        self.db_cursor.execute(sql, params)

        for row in self.db_cursor:
            yield row

#            self.last_regexp_match = None

    @staticmethod
    def normalize_ipaddress(ip_string):
        try:
            packed = socket.inet_pton(socket.AF_INET, ip_string)
            ip_string = socket.inet_ntop(socket.AF_INET, packed)

            ipv6_words = None
            ipv4_int = struct.unpack("!L", packed)[0]
        except socket.error:
            packed = socket.inet_pton(socket.AF_INET6, ip_string)
            ip_string = socket.inet_ntop(socket.AF_INET6, packed)

            ipv6_words = struct.unpack("!8H", packed)
            ipv4_int = None

        return ip_string, ipv4_int, ipv6_words




class WordDetector(object):

    _wdict = {}

    def add_word(self, wid, word):
        prefix = word[:3]
        if prefix not in self._wdict:
            self._wdict[prefix] = [(wid,word),]
        else:
            self._wdict[prefix].append((wid,word),)

    def find_words(self, text):
        found_words = set()
        found_words.add(0)
        s = text.lower()
        for x in range(0, len(s) - 2):
            word_infos = self._wdict.get(s[x:x+3], ())
            found_words.update(wi[0] for wi in word_infos if s.startswith(wi[1], x))
        return found_words


import os

from .base import cached_property
from .base import UdgerBase

import gzip
import io
import tempfile
import urllib.request


class UdgerDownloader(object):

    udger_data_file = 'udgerdb_v4.dat.gz'
    download_url = 'http://data.udger.com/'

    def __init__(self, client_key, data_dir=None):
        self.client_key = client_key
        self.data_dir = data_dir or tempfile.gettempdir()

    def download(self):
        url = UdgerDownloader.download_url + self.client_key + '/' + UdgerDownloader.udger_data_file
        response = urllib.request.urlopen(url)
        compressed_file = io.BytesIO(response.read())
        decompressed_file = gzip.GzipFile(fileobj=compressed_file)

        with open(self.new_filename, 'wb') as outfile:
            outfile.write(decompressed_file.read())
        os.rename(self.new_filename, os.path.join(self.data_dir, UdgerBase.db_filename))

    @cached_property
    def new_filename(self):
        return os.path.join(self.data_dir, UdgerBase.db_filename + '.new')



from os.path import dirname, join
from setuptools import setup, find_packages

from udger import __version__


setup(
    name='udger-v4',
    version=__version__,
    license='BSD',
    author='The Udger Team',
    author_email='info@udger.com',
    description="Fast and reliable User Agent parser and IP classifier for Python",
    long_description_content_type='text/markdown',
    url='https://github.com/udger/udger-python',
    packages=find_packages(
        exclude=('udger.tests',),
    ),
    platforms='any',
    zip_safe=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries",
    ],
)




def join_sql_columns(columns_dict, set_index=None):
    def columns():
        for col_name, col_expression in columns_dict.items():
            if set_index is not None:
                col_expression = col_expression[set_index]

            if col_expression == col_name:
                yield col_name
            else:
                yield '{0} AS {1}'.format(col_expression or 'NULL', col_name)

    return ',\n            '.join(columns())

class Queries(object):
    client_columns = {
        'client_id':                   (None,                          'client_id'),
        'class_id':                    (None,                          'class_id'),

        'ua_class':                    ('"Crawler"',                   'client_classification'),
        'ua_class_code':               ('"crawler"',                   'client_classification_code'),
        'ua':                          ('name',                        'name'),
        'ua_engine':                   (None,                          'engine'),
        'ua_version':                  ('ver',                         None),
        'ua_version_major':            ('ver_major',                   None),
        'crawler_last_seen':           ('last_seen',                   None),
        'crawler_respect_robotstxt':   ('respect_robotstxt',           None),
        'crawler_category':            ('crawler_classification',      None),
        'crawler_category_code':       ('crawler_classification_code', None),
        'ua_uptodate_current_version': (None,                          'uptodate_current_version'),
        'ua_family':                   ('family',                      'name'),
        'ua_family_code':              ('family_code',                 'name_code'),
        'ua_family_homepage':          ('family_homepage',             'homepage'),
        'ua_family_icon':              ('family_icon',                 'icon'),
        'ua_family_icon_big':          (None,                          'icon_big'),
        'ua_family_vendor':            ('vendor',                      'vendor'),
        'ua_family_vendor_code':       ('vendor_code',                 'vendor_code'),
        'ua_family_vendor_homepage':   ('vendor_homepage',             'vendor_homepage'),
        'ua_family_info_url':          ('"https://udger.com/resources/ua-list/bot-detail?bot=" || '
                                        'REPLACE(family, " ", "%20") || "#id" || udger_crawler_list.id',
                                        '"https://udger.com/resources/ua-list/browser-detail?browser=" || '
                                        'REPLACE(name, " ", "%20")'),
    }

    crawler_sql = """
        SELECT
            %s
        FROM
            udger_crawler_list
        LEFT JOIN
            udger_crawler_class ON udger_crawler_class.id = udger_crawler_list.class_id
        WHERE
            ua_string = ?
    """ % join_sql_columns(client_columns, 0)

    client_sql = """
        SELECT
            %s
        FROM
            udger_client_regex ur
        JOIN
            udger_client_list ON udger_client_list.id = ur.client_id
        JOIN
            udger_client_class ON udger_client_class.id = udger_client_list.class_id
        WHERE
            ur.rowid=?
    """ % join_sql_columns(client_columns, 1)

    os_columns = {
        'os_family':                  'family',
        'os_family_code':             'family_code',
        'os':                         'name',
        'os_code':                    'name_code',
        'os_homepage':                'homepage',
        'os_icon':                    'icon',
        'os_icon_big':                'icon_big',
        'os_family_vendor':           'vendor',
        'os_family_vendor_code':      'vendor_code',
        'os_family_vendor_homepage':  'vendor_homepage',
        'os_info_url':                '"https://udger.com/resources/ua-list/os-detail?os=" || '
                                      'REPLACE(name, " ", "%20")',
    }

    os_sql = """
        SELECT
            %s
        FROM
            udger_os_regex ur
        JOIN
            udger_os_list ON udger_os_list.id = ur.os_id
        WHERE
            ur.rowid=?
    """ % join_sql_columns(os_columns)

    client_os_sql = """
        SELECT
            %s
        FROM
            udger_client_os_relation
        JOIN
            udger_os_list ON udger_os_list.id = udger_client_os_relation.os_id
        WHERE
            client_id = ?
    """ % join_sql_columns(os_columns)

    device_columns = {
        'device_class':           'name',
        'device_class_code':      'name_code',
        'device_class_icon':      'icon',
        'device_class_icon_big':  'icon_big',
        'device_class_info_url':  '"https://udger.com/resources/ua-list/device-detail?device=" || '
                                  'REPLACE(name, " ", "%20")',
    }

    device_sql = """
        SELECT
            %s
        FROM
            udger_deviceclass_regex ur
        JOIN
            udger_deviceclass_list ON udger_deviceclass_list.id = ur.deviceclass_id
        WHERE
            ur.rowid=?
    """ % join_sql_columns(device_columns)

    client_class_sql = """
        SELECT
            %s
        FROM
            udger_deviceclass_list
        JOIN
            udger_client_class ON udger_client_class.deviceclass_id = udger_deviceclass_list.id
        WHERE
            udger_client_class.id = ?
    """ % join_sql_columns(device_columns)

    devicename_sql = """
        SELECT
            id AS regex_id,
            regstring
        FROM
            udger_devicename_regex
        WHERE
            os_family_code = ? AND
            os_code IN ('-all-', ?)
        ORDER BY
            sequence
    """

    marketname_columns = {
        'device_marketname':       'marketname',
        'device_brand':            'brand',
        'device_brand_code':       'brand_code',
        'device_brand_homepage':   'brand_url',
        'device_brand_icon':       'icon',
        'device_brand_icon_big':   'icon_big',
        'device_brand_info_url':   '"https://udger.com/resources/ua-list/devices-brand-detail?brand=" || '
                                   'REPLACE(brand_code, " ", "%20")',
    }

    marketname_sql = """
        SELECT
            %s
        FROM
            udger_devicename_list
        JOIN
            udger_devicename_brand ON udger_devicename_brand.id = udger_devicename_list.brand_id
        WHERE
            regex_id = ? AND code = ?
        LIMIT 1
    """ % join_sql_columns(marketname_columns)

    ip_columns = {
        'ip_classification':              'ip_classification',
        'ip_classification_code':         'ip_classification_code',
        'ip_last_seen':                   'ip_last_seen',
        'ip_hostname':                    'ip_hostname',
        'ip_country':                     'ip_country',
        'ip_country_code':                'ip_country_code',
        'ip_city':                        'ip_city',
        'crawler_name':                   'name',
        'crawler_ver':                    'ver',
        'crawler_ver_major':              'ver_major',
        'crawler_family':                 'family',
        'crawler_family_code':            'family_code',
        'crawler_family_homepage':        'family_homepage',
        'crawler_family_vendor':          'vendor',
        'crawler_family_vendor_code':     'vendor_code',
        'crawler_family_vendor_homepage': 'vendor_homepage',
        'crawler_family_icon':            'family_icon',
        'crawler_family_info_url':        '"https://udger.com/resources/ua-list/bot-detail?bot=" || '
                                          'REPLACE(family, " ", "%20") || "#id" || udger_crawler_list.id',
        'crawler_last_seen':              'last_seen',
        'crawler_category':               'crawler_classification',
        'crawler_category_code':          'crawler_classification_code',
        'crawler_respect_robotstxt':      'respect_robotstxt',
    }

    ip_sql = """
        SELECT
            %s
        FROM
            udger_ip_list
        JOIN
            udger_ip_class ON udger_ip_class.id=udger_ip_list.class_id
        LEFT JOIN
            udger_crawler_list ON udger_crawler_list.id=udger_ip_list.crawler_id
        LEFT JOIN
            udger_crawler_class ON udger_crawler_class.id=udger_crawler_list.class_id
        WHERE
            ip = ?
        ORDER BY
            sequence
    """ % join_sql_columns(ip_columns)

    datacenter_columns = {
        'datacenter_name':       'name',
        'datacenter_name_code':  'name_code',
        'datacenter_homepage':   'homepage',
    }

    datacenter_sql = """
        SELECT
            %s
        FROM
            udger_datacenter_range
        JOIN
            udger_datacenter_list ON udger_datacenter_range.datacenter_id = udger_datacenter_list.id
        WHERE
            iplong_from <= ? AND iplong_to >= ?
    """ % join_sql_columns(datacenter_columns)

    datacenter6_sql = """
        SELECT
            %s
        FROM
            udger_datacenter_range6
        JOIN
            udger_datacenter_list ON udger_datacenter_range6.datacenter_id = udger_datacenter_list.id
        WHERE
            iplong_from0 <= ? AND iplong_from1 <= ? AND iplong_from2 <= ? AND iplong_from3 <= ? AND
            iplong_from4 <= ? AND iplong_from5 <= ? AND iplong_from6 <= ? AND iplong_from7 <= ? AND
            iplong_to0 >= ? AND iplong_to1 >= ? AND iplong_to2 >= ? AND iplong_to3 >= ? AND
            iplong_to4 >= ? AND iplong_to5 >= ? AND iplong_to6 >= ? AND iplong_to7 >= ?
    """ % join_sql_columns(datacenter_columns)

    client_ch_regex_columns = {
        'class_id': 'class_id',
        'client_id': 'client_id',
        'regstring': 'regstring',
        'name': 'name',
        'name_code': 'name_code',
        'homepage': 'homepage',
        'icon': 'icon',
        'icon_big': 'icon_big',
        'engine': 'engine',
        'vendor': 'vendor',
        'vendor_code': 'vendor_code',
        'vendor_homepage': 'vendor_homepage',
        'uptodate_current_version': 'uptodate_current_version',
        'client_classification': 'client_classification',
        'client_classification_code ': 'client_classification_code ',
        'ua_family_info_url' : '"https://udger.com/resources/ua-list/browser-detail?browser=" || REPLACE(name, " ", "%20")'
    }

    client_ch_regex_sql = """
        SELECT
            %s
        FROM
            udger_client_ch_regex
        JOIN
            udger_client_list ON udger_client_list.id=udger_client_ch_regex.client_id
        JOIN
            udger_client_class ON udger_client_class.id=udger_client_list.class_id
        WHERE
            udger_client_ch_regex.mobile = ?
        ORDER BY
            sequence ASC
    """ % join_sql_columns(client_ch_regex_columns)

    os_ch_regex_columns = {
        'os_id': 'os_id',
        'regstring': 'regstring',
        'family': 'family',
        'family_code': 'family_code',
        'name,name_code': 'name,name_code',
        'homepage,icon': 'homepage,icon',
        'icon_big,vendor': 'icon_big,vendor',
        'vendor_code': 'vendor_code',
        'vendor_homepage': 'vendor_homepage',
        'os_info_url': '"https://udger.com/resources/ua-list/os-detail?os=" || REPLACE(name, " ", "%20")'
    }

    os_ch_regex_sql = """
        SELECT  
            %s
        FROM  
            udger_os_ch_regex  
        JOIN  
            udger_os_list ON udger_os_list.id=udger_os_ch_regex.os_id  
        WHERE  
            udger_os_ch_regex.version = ? OR udger_os_ch_regex.version = '-all-'  
        ORDER BY sequence ASC
    """ % join_sql_columns(os_ch_regex_columns)

    device_name_ch_regex_sql = """
        SELECT
            id  as id
        FROM
            udger_devicename_regex
        WHERE
            (os_family_code = ? AND os_code = '-all-')
            OR
            (os_family_code = ? AND os_code = ?)
        ORDER BY
            sequence
        LIMIT 0,1
    """

    device_name_list_ch_columns = {
        'marketname': 'marketname',
        'brand_code': 'brand_code',
        'brand': 'brand',
        'brand_url': 'brand_url',
        'icon': 'icon',
        'icon_big': 'icon_big',
        'deviceclass_id': 'deviceclass_id',
        'brand_info_url': '"https://udger.com/resources/ua-list/devices-brand-detail?brand=" || REPLACE(brand_code, " ", "%20")'
    }

    device_name_list_ch_sql = """
        SELECT  
            %s
        FROM  
            udger_devicename_list  
        JOIN  
            udger_devicename_brand ON udger_devicename_brand.id=udger_devicename_list.brand_id  
        WHERE  
            regex_id = ? AND code = ?
    """ % join_sql_columns(device_name_list_ch_columns)

    device_class_ch_sql = """
        SELECT  
            %s 
        FROM  
            udger_deviceclass_list  
        WHERE  
            id = ?
    """ % join_sql_columns(device_columns)

    device_class_by_mobile_ch_sql = """
        SELECT  
            %s 
        FROM  
            udger_deviceclass_list  
        JOIN  
            udger_deviceclass_ch ON udger_deviceclass_list.id=udger_deviceclass_ch.device_id  
        WHERE  
            udger_deviceclass_ch.mobile = ?
    """ % join_sql_columns(device_columns)

